@@ -17,8 +17,8 @@
 //
 // Alternative licensing terms are available from the licensor.
 // For commercial licensing, see <https://www.artifex.com/> or contact
-// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
-// CA 94945, U.S.A., +1(415)492-9861, for further information.
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
 
 package com.artifex.mupdf.fitz;
 
@@ -105,7 +105,7 @@ public class PDFObject implements Iterable<PDFObject>
 	}
 
 	private native PDFObject getArray(int index);
-	private native PDFObject getDictionary(String name);
+	private native PDFObject getDictionary(String name, boolean inheritable);
 	private native PDFObject getDictionaryKey(int index);
 
 	public PDFObject get(int index) {
@@ -113,11 +113,19 @@ public class PDFObject implements Iterable<PDFObject>
 	}
 
 	public PDFObject get(String name) {
-		return getDictionary(name);
+		return getDictionary(name, false);
 	}
 
 	public PDFObject get(PDFObject name) {
-		return getDictionary(name != null ? name.asName() : null);
+		return getDictionary(name != null ? name.asName() : null, false);
+	}
+
+	public PDFObject getInheritable(String name) {
+		return getDictionary(name, true);
+	}
+
+	public PDFObject getInheritable(PDFObject name) {
+		return getDictionary(name != null ? name.asName() : null, true);
 	}
 
 	private native void putArrayBoolean(int index, boolean b);
@@ -131,6 +139,9 @@ public class PDFObject implements Iterable<PDFObject>
 	private native void putDictionaryStringFloat(String name, float f);
 	private native void putDictionaryStringString(String name, String str);
 	private native void putDictionaryStringPDFObject(String name, PDFObject obj);
+	private native void putDictionaryStringRect(String name, Rect r);
+	private native void putDictionaryStringMatrix(String name, Matrix m);
+	private native void putDictionaryStringDate(String name, long secs);
 
 	private native void putDictionaryPDFObjectBoolean(PDFObject name, boolean b);
 	private native void putDictionaryPDFObjectInteger(PDFObject name, int i);
@@ -181,6 +192,18 @@ public class PDFObject implements Iterable<PDFObject>
 		putDictionaryStringPDFObject(name, obj);
 	}
 
+	public void put(String name, Rect r) {
+		putDictionaryStringRect(name, r);
+	}
+
+	public void put(String name, Matrix m) {
+		putDictionaryStringMatrix(name, m);
+	}
+
+	public void put(String name, Date time) {
+		putDictionaryStringDate(name, time.getTime());
+	}
+
 	public void put(PDFObject name, boolean b) {
 		putDictionaryPDFObjectBoolean(name, b);
 	}
