@@ -12,7 +12,7 @@ import textwrap
 
 
 class WindowsVS:
-    '''
+    r'''
     Windows only. Finds locations of Visual Studio command-line tools. Assumes
     VS2019-style paths.
 
@@ -34,7 +34,8 @@ class WindowsVS:
         '''
         Args:
             year:
-                None or, for example, `2019`.
+                None or, for example, `2019`. If None we use environment
+                variable WDEV_VS_YEAR if set.
             grade:
                 None or, for example, one of:
 
@@ -42,11 +43,24 @@ class WindowsVS:
                 * `Professional`
                 * `Enterprise`
 
+                If None we use environment variable WDEV_VS_GRADE if set.
             version:
-                None or, for example: `14.28.29910`
+                None or, for example: `14.28.29910`. If None we use environment
+                variable WDEV_VS_VERSION if set.
             cpu:
                 None or a `WindowsCpu` instance.
         '''
+        def default(value, name):
+            if value is None:
+                name2 = f'WDEV_VS_{name.upper()}'
+                value = os.environ.get(name2)
+                if value is not None:
+                    _log(f'Setting {name} from environment variable {name2}: {value!r}')
+            return value
+        year = default(year, 'year')
+        grade = default(grade, 'grade')
+        version = default(version, 'version')
+
         if not cpu:
             cpu = WindowsCpu()
 
@@ -87,7 +101,7 @@ class WindowsVS:
 
         # Find vcvars.bat.
         #
-        vcvars = f'{directory}\\VC\Auxiliary\\Build\\vcvars{cpu.bits}.bat'
+        vcvars = f'{directory}\\VC\\Auxiliary\\Build\\vcvars{cpu.bits}.bat'
         assert os.path.isfile( vcvars), f'No match for: {vcvars}'
 
         # Find cl.exe.
@@ -126,8 +140,6 @@ class WindowsVS:
                     csc = os.path.join(dirpath, filename)
                     #_log(f'{csc=}')
                     #break
-            #if csc:
-            #    break
 
         self.cl = cl
         self.devenv = devenv
@@ -201,8 +213,8 @@ class WindowsCpu:
 
 class WindowsPython:
     '''
-    Experimental. Windows only. Information about installed Python with
-    specific word size and version. Defaults to the currently-running Python.
+    Windows only. Information about installed Python with specific word size
+    and version. Defaults to the currently-running Python.
 
     Members:
 
@@ -295,7 +307,7 @@ def _cpu_name():
     '''
     Returns `x32` or `x64` depending on Python build.
     '''
-    _log(f'sys.maxsize={hex(sys.maxsize)}')
+    #log(f'sys.maxsize={hex(sys.maxsize)}')
     return f'x{32 if sys.maxsize == 2**31 - 1 else 64}'
 
 
