@@ -69,16 +69,19 @@ def build_swig(
     assert language in ('python', 'csharp')
     # Find version of swig. (We use quotes around <swig> to make things work on
     # Windows.)
-    t = jlib.system( f'"{swig_command}" -version', out='return', verbose=1)
-    jlib.log('SWIG version info:\n========\n{t}\n========')
+    e, swig_location = jlib.system( f'which "{swig_command}"', raise_errors=0, out='return', verbose=0)
+    if e == 0:
+        jlib.log(f'{swig_location=}')
+    t = jlib.system( f'"{swig_command}" -version', out='return', verbose=0)
+    jlib.log1('SWIG version info:\n========\n{t}\n========')
     m = re.search( 'SWIG Version ([0-9]+)[.]([0-9]+)[.]([0-9]+)', t)
     assert m
     swig_major = int( m.group(1))
-    jlib.system( f'which "{swig_command}"')
+    jlib.log(f'{m.group()}')
 
     # Create a .i file for SWIG.
     #
-    common = f'''
+    common = textwrap.dedent(f'''
             #include <stdexcept>
 
             #include "mupdf/functions.h"
@@ -97,6 +100,7 @@ def build_swig(
             #endif
 
             '''
+            )
     if language == 'csharp':
         common += textwrap.dedent(f'''
                 /* This is required otherwise compiling the resulting C++ code
@@ -820,6 +824,12 @@ def build_swig(
             %ignore {rename.ll_fn('Memento_vasprintf')};
             %ignore {rename.fn('Memento_vasprintf')};
 
+            // These appear to be not present in Windows debug builds.
+            %ignore fz_assert_lock_held;
+            %ignore fz_assert_lock_not_held;
+            %ignore fz_lock_debug_lock;
+            %ignore fz_lock_debug_unlock;
+
             %ignore Memento_cpp_new;
             %ignore Memento_cpp_delete;
             %ignore Memento_cpp_new_array;
@@ -844,8 +854,36 @@ def build_swig(
             // This appears to allow python to call fns taking an int64_t.
             %include "stdint.i"
 
+            /*
+            This is only documented for Ruby, but is mentioned for Python at
+            https://sourceforge.net/p/swig/mailman/message/4867286/.
+
+            It makes the Python wrapper for `FzErrorBase` inherit Python's
+            `Exception` instead of `object`, which in turn means it can be
+            caught in Python with `except Exception as e: ...` or similar.
+
+            Note that while it will have the underlying C++ class's `what()`
+            method, this is not used by the `__str__()` and `__repr__()`
+            methods. Instead:
+
+                `__str__()` appears to return a tuple of the constructor args
+                that were originally used to create the exception object with
+                `PyObject_CallObject(class_, args)`.
+
+                `__repr__()` returns a SWIG-style string such as
+                `<texcept.MyError; proxy of <Swig Object of type 'MyError *' at
+                0xb61ebfabc00> >`.
+
+            We explicitly overwrite `__str__()` to call `what()`.
+            */
+            %feature("exceptionclass")  FzErrorBase;
+
             %{{
-            {common}
+            ''')
+
+    text += common
+
+    text += textwrap.dedent(f'''
             %}}
 
             %include exception.i
@@ -853,6 +891,8 @@ def build_swig(
             %include carrays.i
             %include cdata.i
             %include std_vector.i
+            %include std_map.i
+
             {"%include argcargv.i" if language=="python" else ""}
 
             %array_class(unsigned char, uchar_array);
@@ -863,8 +903,12 @@ def build_swig(
             {{
                 %template(vectoruc) vector<unsigned char>;
                 %template(vectori) vector<int>;
+                %template(vectorf) vector<float>;
+                %template(vectord) vector<double>;
                 %template(vectors) vector<std::string>;
+                %template(map_string_int) map<std::string, int>;
                 %template(vectorq) vector<{rename.namespace_class("fz_quad")}>;
+                %template(vector_search_page2_hit) vector<fz_search_page2_hit>;
             }};
 
             // Make sure that operator++() gets converted to __next__().
@@ -892,41 +936,8 @@ def build_swig(
             %array_functions(float, floats);
             ''')
 
-    text += textwrap.dedent(f'''
-            %exception
-            {{
-                try
-                {{
-                    $action
-                }}
-                catch( std::exception& e)
-                {{
-                    if (g_mupdf_trace_exceptions)
-                    {{
-                        std::cerr
-                                #ifndef _WIN32
-                                << __PRETTY_FUNCTION__ << ": "
-                                #endif
-                                << "Converting C++ std::exception into Python exception: " << e.what()
-                                << "\\n";
-                    }}
-                    SWIG_exception( SWIG_RuntimeError, e.what());
-                }}
-                catch(...)
-                {{
-                    if (g_mupdf_trace_exceptions)
-                    {{
-                        std::cerr
-                                #ifndef _WIN32
-                                << __PRETTY_FUNCTION__ << ": "
-                                #endif
-                                << "Converting unknown C++ exception into Python exception."
-                                << "\\n";
-                    }}
-                    SWIG_exception( SWIG_RuntimeError, "Unknown exception");
-                }}
-            }}
-            ''')
+    if language == 'python':
+        text += generated.swig_python_exceptions.getvalue()
 
     text += textwrap.dedent(f'''
             // Ensure SWIG handles OUTPUT params.
@@ -1030,6 +1041,7 @@ def build_swig(
         # tuples.
         #
         text += generated.swig_python
+        text += generated.swig_python_set_error_classes.getvalue()
 
         def set_class_method(struct, fn):
             return f'{rename.class_(struct)}.{rename.method(struct, fn)} = {fn}'
@@ -1504,6 +1516,71 @@ def build_swig(
                         self._out = out
                     return FzDocumentWriter__init__0(self, *args)
                 FzDocumentWriter.__init__ = FzDocumentWriter__init__1
+
+                # Create class derived from
+                # fz_install_load_system_font_funcs_args class wrapper with
+                # overrides of the virtual functions to allow calling of Python
+                # callbacks.
+                #
+                class fz_install_load_system_font_funcs_args3({rename.class_('fz_install_load_system_font_funcs_args')}2):
+                    """
+                    Class derived from Swig Director class
+                    fz_install_load_system_font_funcs_args2, to allow
+                    implementation of fz_install_load_system_font_funcs with
+                    Python callbacks.
+                    """
+                    def __init__(self, f=None, f_cjk=None, f_fallback=None):
+                        super().__init__()
+
+                        self.f3 = f
+                        self.f_cjk3 = f_cjk
+                        self.f_fallback3 = f_fallback
+
+                        self.use_virtual_f(True if f else False)
+                        self.use_virtual_f_cjk(True if f_cjk else False)
+                        self.use_virtual_f_fallback(True if f_fallback else False)
+
+                    def ret_font(self, font):
+                        if font is None:
+                            return None
+                        elif isinstance(font, {rename.class_('fz_font')}):
+                            return ll_fz_keep_font(font.m_internal)
+                        elif isinstance(font, fz_font):
+                            return font
+                        else:
+                            assert 0, f'Expected FzFont or fz_font, but fz_install_load_system_font_funcs() callback returned {{type(font)=}}'
+
+                    def f(self, ctx, name, bold, italic, needs_exact_metrics):
+                        font = self.f3(name, bold, italic, needs_exact_metrics)
+                        return self.ret_font(font)
+
+                    def f_cjk(self, ctx, name, ordering, serif):
+                        font = self.f_cjk3(name, ordering, serif)
+                        return self.ret_font(font)
+
+                    def f_fallback(self, ctx, script, language, serif, bold, italic):
+                        font = self.f_fallback3(script, language, serif, bold, italic)
+                        return self.ret_font(font)
+
+                # We store the most recently created
+                # fz_install_load_system_font_funcs_args in this global so that
+                # it is not cleaned up by Python.
+                g_fz_install_load_system_font_funcs_args = None
+
+                def fz_install_load_system_font_funcs(f=None, f_cjk=None, f_fallback=None):
+                    """
+                    Python override for MuPDF
+                    fz_install_load_system_font_funcs() using Swig Director
+                    support. Python callbacks are not passed a `ctx` arg, and
+                    can return None, a mupdf.fz_font or a mupdf.FzFont.
+                    """
+                    global g_fz_install_load_system_font_funcs_args
+                    g_fz_install_load_system_font_funcs_args = fz_install_load_system_font_funcs_args3(
+                            f,
+                            f_cjk,
+                            f_fallback,
+                            )
+                    fz_install_load_system_font_funcs2(g_fz_install_load_system_font_funcs_args)
                 ''')
 
         # Add __iter__() methods for all classes with begin() and end() methods.
@@ -1641,6 +1718,10 @@ def build_swig(
         def make_command( module, cpp, swig_i):
             cpp = os.path.relpath( cpp)
             swig_i = os.path.relpath( swig_i)
+            # We need to predefine MUPDF_FITZ_HEAP_H to disable parsing of
+            # include/mupdf/fitz/heap.h. Otherwise swig's preprocessor seems to
+            # ignore #undef's in include/mupdf/fitz/heap-imp.h then complains
+            # about redefinition of macros in include/mupdf/fitz/heap.h.
             command = (
                     textwrap.dedent(
                     f'''
@@ -1655,10 +1736,12 @@ def build_swig(
                         -outdir {os.path.relpath(build_dirs.dir_so)}
                         -o {cpp}
                         -includeall
+                        {os.environ.get('XCXXFLAGS', '')}
                         -I{os.path.relpath(build_dirs.dir_mupdf)}/platform/python/include
                         -I{os.path.relpath(include1)}
                         -I{os.path.relpath(include2)}
                         -ignoremissing
+                        -DMUPDF_FITZ_HEAP_H
                         {swig_i}
                     ''').strip().replace( '\n', "" if state_.windows else " \\\n")
                     )
@@ -1759,6 +1842,7 @@ def build_swig(
                     -I{os.path.relpath(include1)}
                     -I{os.path.relpath(include2)}
                     -ignoremissing
+                    -DMUPDF_FITZ_HEAP_H
                     {os.path.relpath(swig_i)}
                 ''').strip().replace( '\n', "" if state_.windows else "\\\n")
                 )
@@ -1780,16 +1864,16 @@ def build_swig(
                 '\\2public override string ToString() { return to_string(); }\n\\1',
                 cs,
                 )
-        jlib.log('{len(cs)=}')
-        jlib.log('{len(cs2)=}')
+        jlib.log1('{len(cs)=}')
+        jlib.log1('{len(cs2)=}')
         assert cs2 != cs, f'Failed to add toString() methods.'
-        jlib.log('{len(generated.swig_csharp)=}')
+        jlib.log1('{len(generated.swig_csharp)=}')
         assert len(generated.swig_csharp)
         cs2 += generated.swig_csharp
-        jlib.log( 'Updating cs2 => {build_dirs.dir_so}/mupdf.cs')
+        jlib.log1( 'Updating cs2 => {build_dirs.dir_so}/mupdf.cs')
         jlib.fs_update(cs2, f'{build_dirs.dir_so}/mupdf.cs')
         #jlib.fs_copy(f'{outdir}/mupdf.cs', f'{build_dirs.dir_so}/mupdf.cs')
-        jlib.log('{rebuilt=}')
+        jlib.log1('{rebuilt=}')
 
     else:
         assert 0
