@@ -17,8 +17,8 @@
 //
 // Alternative licensing terms are available from the licensor.
 // For commercial licensing, see <https://www.artifex.com/> or contact
-// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
-// CA 94945, U.S.A., +1(415)492-9861, for further information.
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
 
 #include "mupdf/fitz.h"
 
@@ -244,7 +244,7 @@ fz_new_document_writer(fz_context *ctx, const char *path, const char *explicit_f
 		else
 			format = NULL;
 	}
-	fz_throw(ctx, FZ_ERROR_GENERIC, "cannot detect document format");
+	fz_throw(ctx, FZ_ERROR_ARGUMENT, "cannot detect document format");
 }
 
 fz_document_writer *
@@ -288,7 +288,7 @@ fz_new_document_writer_with_output(fz_context *ctx, fz_output *out, const char *
 		return fz_new_docx_writer_with_output(ctx, out, options);
 #endif
 
-	fz_throw(ctx, FZ_ERROR_GENERIC, "unknown output document format: %s", format);
+	fz_throw(ctx, FZ_ERROR_ARGUMENT, "unknown output document format: %s", format);
 }
 
 fz_document_writer *
@@ -334,7 +334,7 @@ fz_begin_page(fz_context *ctx, fz_document_writer *wri, fz_rect mediabox)
 	if (!wri)
 		return NULL;
 	if (wri->dev)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "called begin page without ending the previous page");
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "called begin page without ending the previous page");
 	wri->dev = wri->begin_page(ctx, wri, mediabox);
 	return wri->dev;
 }
